package org.jcg.zheng.demo.async;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.concurrent.Callable;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import java.util.concurrent.ForkJoinPool;

import org.jcg.zheng.demo.async.tasks.CallableTask;
import org.jcg.zheng.demo.async.tasks.ChainOperations_PrimitiveType;
import org.jcg.zheng.demo.async.tasks.RunnableTask;
import org.jcg.zheng.demo.async.tasks.SupplierTask;
import org.junit.Test;

public class CompletableFutureTest extends TestBase {
    private Runnable runableTask = new RunnableTask();
    private Callable<String> callableTask = new CallableTask();
    private SupplierTask supplierTask = new SupplierTask();

    @Test
    public void complete() throws InterruptedException, ExecutionException {
        CompletableFuture<String> future = new CompletableFuture<>();
        Executors.newCachedThreadPool().submit(callableTask);
        future.complete("Hello");
        assertEquals("Hello", future.get());

    }

    @Test
    public void get_block_Main_thread() throws InterruptedException, ExecutionException {
        CompletableFuture<Integer> future = CompletableFuture.supplyAsync(supplierTask);
        Integer tt = future.get();

        assertTrue(tt.intValue() > 0);
        assertTrue(future.isDone());
        future.thenAccept(d -> printNum(d)); // print from main thread

    }

    @Test
    public void runAsync_for_runable_block_on_get()
            throws InterruptedException, ExecutionException {
        // this is same as Future
        CompletableFuture<Void> future = CompletableFuture.runAsync(runableTask);
        future.get();
        assertTrue(future.isDone());
    }


    @Test
    public void non_blocking_thread_execution() throws InterruptedException {
        CompletableFuture<Integer> future = CompletableFuture.supplyAsync(supplierTask);
        future.thenAccept(d -> printNum(d));// print it from ForkJoinPool.commonPool-worker-#
        assertFalse(future.isDone());
    }

    @Test
    public void non_blocking_thread_execution_chain() throws InterruptedException {
        // print it from ForkJoinPool.commonPool-worker-#
        CompletableFuture.supplyAsync(supplierTask).thenAccept(d -> printNum(d));

    }

    @Test
    public void supplyAsync_thenAccept_Separate() throws InterruptedException {
        CompletableFuture<Integer> future = CompletableFuture.supplyAsync(supplierTask);
        future.thenAccept(d -> printNum(d));// print it from ForkJoinPool.commonPool-worker-#
        assertFalse(future.isDone());
    }

    @Test
    public void supplyAsync_thenAccept_chain() throws InterruptedException {
        // print it from ForkJoinPool.commonPool-worker-#
        CompletableFuture.supplyAsync(supplierTask).thenAccept(d -> printNum(d));

    }

    @Test
    public void chain_theApply_with_lambda() {
        CompletableFuture<String> future = CompletableFuture.completedFuture("Done")
                .thenApply(s -> {
                    return s.toUpperCase();
                });

        assertEquals("DONE", future.getNow(null));
        assertTrue(future.isDone());
    }

    @Test
    public void change_threadpool_to_user_worker_pool() {
        // ForkJoinPool-1-worker-19
        ForkJoinPool fjp = new ForkJoinPool(10);
        CompletableFuture<Integer> future = CompletableFuture.supplyAsync(supplierTask, fjp);
        future.thenAccept(d -> printNum(d));
        assertFalse(future.isDone());

    }

    @Test
    public void chain_thenAccept_with_thenRun() {
        CompletableFuture<Integer> future = CompletableFuture.supplyAsync(supplierTask);

        future.thenAccept(d -> printNum(d)).thenRun(() -> logger.info("Do step 1"))
                .thenRun(() -> logger.info("Do step 2"));
        assertFalse(future.isDone());
    }

    @Test
    public void chain_thenApply_thenAccept() {
        CompletableFuture<Integer> future = CompletableFuture.supplyAsync(supplierTask);
        future.thenApply(old -> old * 2).thenApply(old -> old + 2).thenAccept(d -> printNum(d));
        assertFalse(future.isDone());
    }

    @Test
    public void thenCompose() throws InterruptedException, ExecutionException {
        CompletableFuture<String> future = CompletableFuture.supplyAsync(() -> "Hello")
                .thenCompose(s -> CompletableFuture.supplyAsync(() -> s + " World!"));
        String helloworld = future.get();
        assertEquals("Hello World!", helloworld);
    }

    @Test
    public void thenCombine() {
        ChainOperations_PrimitiveType scP = new ChainOperations_PrimitiveType();
        scP.doubleNum(2).thenCombine(scP.doubleNum(4), (a, b) -> a + b)
                .thenAccept(d -> printNum(d));
        logger.info("It should print 12 =2 * 2 + 4 * 4");

    }

}
