package org.jcg.zheng.demo.async;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.jcg.zheng.demo.async.data.DataUtils;
import org.jcg.zheng.demo.async.tasks.CallableTask;
import org.jcg.zheng.demo.async.tasks.RunnableTask;
import org.junit.Test;

public class FutureTest extends TestBase {

    private ExecutorService service = Executors.newFixedThreadPool(10);
    private Runnable runableTask = new RunnableTask();
    private Callable<String> callableTask = new CallableTask();

    @Test
    public void get_block_runnable() throws InterruptedException, ExecutionException {
        Future future = service.submit(runableTask);
        assertFalse(future.isDone());

        future.get();

        logger.info(DataUtils.MAIN_THREAD_RESUMES);
        assertTrue(future.isDone());
    }

    @Test
    public void concurrent_with_runnable() throws InterruptedException {
        Future future = service.submit(runableTask);
        assertFalse(future.isDone());
    }

    @Test
    public void concurrent_with_runnable_no_blocking_with_more_threads()
            throws InterruptedException {
        List<Future> futures = new ArrayList<>();
        for (int i = 0; i < 5; i++) {
            futures.add(service.submit(runableTask));
        }

        futures.forEach(f -> assertFalse(f.isDone()));
    }

    @Test
    public void get_block_callable() throws InterruptedException, ExecutionException {
        Future<String> future = service.submit(callableTask);

        String name = future.get();

        logger.info(DataUtils.MAIN_THREAD_RESUMES + " with results from async task =" + name);
        assertTrue(future.isDone());
    }

    @Test(expected = java.util.concurrent.TimeoutException.class)
    public void get_with_timeout_block_callable_exceed_timeout()
            throws InterruptedException, ExecutionException, TimeoutException {
        Future<String> future = service.submit(callableTask);

        String name = future.get(3, TimeUnit.SECONDS);
    }

    @Test
    public void get_with_timeout_block_callable_within_timeout_limit()
            throws InterruptedException, ExecutionException, TimeoutException {
        Future<String> future = service.submit(callableTask);

        String name = future.get(7, TimeUnit.SECONDS);
        assertTrue(future.isDone());
        logger.info(DataUtils.MAIN_THREAD_RESUMES + " with results from async task =" + name);
    }

    @Test
    public void it_cancelled_before_async_thread_starts() {
        Future future = service.submit(runableTask);
        assertFalse(future.isDone());

        boolean cancelStatus = future.cancel(true);

        assertTrue(future.isCancelled());
        assertTrue(cancelStatus);
    }

    @Test
    public void it_aborted_after_async_starts_but_before_finish() {
        Future future = service.submit(runableTask);
        assertFalse(future.isDone());

        try {
            TimeUnit.SECONDS.sleep(2);
            boolean cancelStatus = future.cancel(true);

            assertTrue(future.isCancelled());
            assertTrue(cancelStatus);

        } catch (InterruptedException e) {
        }

    }

    @Test
    public void it_failed_cancel_after_async_thread_finish() {
        Future future = service.submit(runableTask);
        assertFalse(future.isDone());

        try {
            TimeUnit.SECONDS.sleep(6);
            boolean cancelStatus = future.cancel(true);

            assertFalse(future.isCancelled());
            assertFalse(cancelStatus);
        } catch (InterruptedException e) {
        }

    }

    @Test
    public void cancel_callable() {
        Future<String> future = service.submit(callableTask);
        assertFalse(future.isDone());

        boolean cancelStatus = future.cancel(true);

        assertTrue(future.isCancelled());
        assertTrue(cancelStatus);
    }

}
